import shutil
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Set, Iterator

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distinfo import DistInfo, RecordsFileConfiguration
from pkm.api.distributions.pth_link import PthLink
from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.packages.site_packages import InstalledPackage
from pkm.api.repositories.repository import AbstractRepository, Repository
from pkm.distributions.executables import Executables
from pkm.utils.commons import NoSuchElementException
from pkm.utils.files import CopyTransaction, is_empty_directory, is_relative_to
from pkm.utils.iterators import first_or_none


class SharedPackagesRepository(AbstractRepository):

    def __init__(self, workspace: Path, base_repository: Repository):
        super().__init__("shared")
        self._workspace = workspace
        self._base_repo = base_repository

    def remove_unused_packages(self, package_users: Iterator[Environment]):
        used_packages: Set[str] = set()
        for package_user in package_users:
            for used_package in package_user.site_packages.installed_packages():
                if (shared_marker := used_package.dist_info.path / "SHARED").exists():
                    used_packages.add(shared_marker.read_text())

        for package_dir in self._workspace.iterdir():
            if not package_dir.is_dir():
                continue

            for version_dir in package_dir.iterdir():
                if not version_dir.is_dir():
                    continue

                for shared_dir in version_dir.iterdir():
                    if not shared_dir.is_dir():
                        continue

                    if str(shared_dir.absolute()) not in used_packages:
                        shutil.rmtree(shared_dir)

                if is_empty_directory(version_dir):
                    version_dir.rmdir()

            if is_empty_directory(package_dir):
                package_dir.rmdir()

    def _do_match(self, dependency: Dependency) -> List[Package]:
        packages_dir = self._workspace / dependency.package_name
        packages = self._base_repo.match(dependency, False)
        return [
            _SharedPackage(p, packages_dir / str(p.version))
            for p in packages
        ]

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        def key(p: Package):
            return 0 if isinstance(p, _SharedPackage) else 1

        packages.sort(key=key)
        return packages


@dataclass
class _SharedPackageArtifact:
    path: Path
    metadata: PackageMetadata
    compatibility_tags: str


class _SharedPackage(Package):

    def __init__(self, package: Package, shared_path: Path):
        self._package = package
        self._shared_path = shared_path

        if shared_path.exists():
            self._artifacts = [
                _SharedPackageArtifact(artifact, PackageMetadata.load(artifact / "dist-info/METADATA"), artifact.name)
                for artifact in shared_path.iterdir()
            ]
        else:
            self._artifacts = []

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._package.descriptor

    def _shared_artifact_for(self, env: Environment) -> Optional[_SharedPackageArtifact]:
        try:
            return max(
                ((a, s) for a in self._artifacts
                 if (s := env.compatibility_tag_score(a.compatibility_tags)) is not None),
                key=lambda it: it[1]
            )[0]
        except ValueError:
            return None

    def _all_dependencies(self, environment: "Environment") -> List["Dependency"]:
        if artifact := self._shared_artifact_for(environment):
            return artifact.metadata.dependencies

        return self._package.dependencies(environment)

    def is_compatible_with(self, env: "Environment") -> bool:
        return self._package.is_compatible_with(env)

    def install_to(
            self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
            editable: bool = False):
        if shared := self._shared_artifact_for(target.env):
            _link_shared(self.descriptor, shared, target)
        else:
            self._package.install_to(target, user_request, editable)
            if shared := _move_to_shared(self._package.descriptor, target, self._shared_path):
                _link_shared(self.descriptor, shared, target)


def _copy_records(root: Path, shared: Path, records: List[Path]) -> List[Path]:
    shared.mkdir(parents=True)
    records_left = []
    for record in records:
        if is_relative_to(record, root):  # the path may lead out of the required source
            shared_path = shared / record.relative_to(root)
            if record.is_dir():
                shared_path.mkdir(exist_ok=True, parents=True)
            elif record.exists():
                shared_path.parent.mkdir(exist_ok=True, parents=True)
                shutil.copy(record, shared_path)
        else:
            records_left.append(record)
    return records_left


def _move_to_shared(package: PackageDescriptor,
                    target: PackageInstallationTarget, shared_path: Path) -> Optional[_SharedPackageArtifact]:
    psname = (package.expected_src_package_name + "-").lower()

    for site in ('purelib', 'platlib'):
        dist_info_path = first_or_none(
            it for it in Path(getattr(target, site)).glob("*.dist-info") if it.name.lower().startswith(psname))

        if dist_info_path:
            break
    else:
        raise NoSuchElementException(f"package: {package} is reported as installed but could not be find inside venv")

    dist_info = DistInfo.load(dist_info_path)

    archive = (dist_info.path / "INSTALLER").read_text().splitlines(keepends=False)
    if len(archive) >= 2 and archive[1].endswith(".whl"):
        compatibility_tags = WheelDistribution.compute_compatibility_tags_of(Path(archive[1]))
        shared_target = shared_path / compatibility_tags
    else:
        return None

    records = list(dist_info.installed_files())

    # filter dist info records:
    records = [
        r for r in records
        if not is_relative_to(r, dist_info.path)]

    # copy site
    shared_purelib = shared_target / site
    records = _copy_records(dist_info.path.parent, shared_purelib, records)

    # copy scripts
    script_entrypoints = {e.name for e in dist_info.load_entrypoints_cfg().entrypoints if e.is_script()}
    bin_path = Path(target.scripts)
    shared_bin = shared_target / "bin"

    # filter script entrypoints - they should be re-generated
    records = [r for r in records if not is_relative_to(r, bin_path) or r.stem not in script_entrypoints]
    records = _copy_records(bin_path, shared_bin, records)

    # copy dist-info
    shutil.copytree(dist_info.path, shared_target / 'dist-info')

    if records:
        warnings.warn(f"{len(records)} unsharable records found in package: {package.name} {package.version}: "
                      f"{', '.join(str(r) for r in records)}")

    # uninstall the original package
    InstalledPackage(dist_info).uninstall()

    return _SharedPackageArtifact(shared_target, dist_info.load_metadata_cfg(), shared_target.name)


def _link_shared(package: PackageDescriptor, shared: _SharedPackageArtifact, target: PackageInstallationTarget):
    with CopyTransaction() as ct:

        package_prefix = f"{package.expected_src_package_name}-{package.version}"

        site_name = "purelib" if (shared.path / "purelib").exists() else "platlib"
        site_path = Path(getattr(target, site_name))

        # first link the site data
        purelib_link = PthLink(site_path / f"{package_prefix}.pth", [shared.path / site_name])
        purelib_link.save()
        ct.touch(purelib_link.path)

        # now create all script entrypoints
        shared_distinfo = DistInfo.load(shared.path / "dist-info", non_standard_name_ok=True)
        bin_dir = Path(target.scripts)
        for entrypoint in shared_distinfo.load_entrypoints_cfg().entrypoints:
            if entrypoint.is_script():
                ct.touch(Executables.generate_for_entrypoint(target.env, entrypoint, bin_dir))

        # then, patch non entrypoint scripts
        for script in (shared.path / 'bin').iterdir():
            target_script = bin_dir / script.name
            Executables.patch_shabang_for_interpreter(script, target_script, target.env.interpreter_path)
            ct.touch(target_script)

        # we are almost done, copy dist-info
        distinfo_path = site_path / f"{package_prefix}.dist-info"
        ct.copy_tree(shared_distinfo.path, distinfo_path, lambda it: it.name != "RECORD")
        shared_marker = (distinfo_path / "SHARED")
        shared_marker.write_text(str(shared.path.absolute()))
        ct.touch(shared_marker)

        # and finally, sign the installation
        record_path = distinfo_path / "RECORD"
        RecordsFileConfiguration.load(record_path).sign_files(ct.copied_files, site_path).save()



import configparser
from abc import ABC, abstractmethod
from copy import copy
from pathlib import Path
from typing import Optional, Dict, Any, List, Sequence, Mapping, Iterator, Callable, TypeVar, Tuple, Type, cast, \
    MutableMapping, Generic, Iterable

from pkm.config import toml
from pkm.utils.commons import UnsupportedOperationException


class Configuration:

    def __init__(
            self, *,
            parent: Optional["Configuration"] = None,
            data: Optional[Dict[str, Any]] = None):
        self._parent = parent
        self._data: Dict[str, Any] = data if data is not None else {}

    def __getitem__(self, item: Sequence[str]) -> Any:
        if isinstance(item, str):
            item = toml.key2path(item)

        r = self._get(item)
        if r is None and self._parent:
            r = self._parent[item]
        return r

    def __contains__(self, item: Sequence[str]):
        return self[item] is not None

    def _get(self, path: Sequence[str]) -> Any:
        r = self._data
        for p in path:
            if r is None:
                return None
            if not isinstance(r, Mapping):
                raise ValueError(f"path: {path} passing through a terminal value {r} at '{p}'")

            r = r.get(p)

        return r

    def _subs_chain(self) -> Iterator["Configuration"]:
        result = []
        r = self
        while r is not None:
            result.append(r)
            r = r._parent

        return reversed(result)

    def collect(self, path: Sequence[str]) -> Any:
        if isinstance(path, str):
            path = toml.key2path(path)

        result = self[path]
        if isinstance(result, List):
            return [it for conf in self._subs_chain() for it in (conf._get(path) or [])]
        elif isinstance(result, Mapping):
            return {k: v for conf in self._subs_chain() for k, v in (conf._get(path).items() or {})}

        return result

    def with_parent(self, new_parent: Optional["Configuration"]) -> "Configuration":
        cp = copy(self)
        cp._parent = new_parent
        return cp

    def items(self) -> Iterable[Tuple[str, Any]]:
        return cast(Iterable[Tuple[str, Any]], self._data.items())


class MutableConfiguration(Configuration, ABC):
    def __setitem__(self, key: Sequence[str], value: Any):
        if isinstance(key, str):
            key = toml.key2path(key)

        self._set(key, value)

    def __delitem__(self, key: Sequence[str]):
        if isinstance(key, str):
            key = toml.key2path(key)

        self._del(key)

    @abstractmethod
    def save_to(self, path: Optional[Path] = None):
        ...

    def save(self) -> bool:
        self.save_to(None)
        return True

    def _set(self, path: Sequence[str], value: Any):
        r = self._data
        for p in path[:-1]:
            if not isinstance(r, MutableMapping):
                raise ValueError(f"path: {path} passing through a terminal value {r} at '{p}'")

            rp = r.get(p)
            if rp is None:
                r[p] = r = {}
            else:
                r = rp

        r[path[-1]] = value

    def get_or_put(self, key: Sequence[str], value_computer: Callable[[], Any]) -> Any:
        if isinstance(key, str):
            key = toml.key2path(key)

        result = self[key]
        if result is None:
            self[key] = result = value_computer()

        return result

    def _del(self, path: Sequence[str]):
        r = self._data
        for p in path[:-1]:
            if r is None:
                return

            if not isinstance(r, Mapping):
                raise ValueError(f"path: {path} passing through a terminal value {r} at '{p}'")

            r = r.get(p)

        if r is not None:
            del r[path[-1]]


_T = TypeVar('_T')
_C1R = TypeVar("_C1R", bound=Callable[[Any], Any])
_C1N = TypeVar("_C1N", bound=Callable[[Any], None])


class _ComputedConfigValue(Generic[_T]):

    def __init__(self, func: Callable[[Any], _T], dependency_keys: Tuple[str, ...]):
        self._func = func
        self._dependency_keys = tuple(toml.key2path(key) for key in dependency_keys)
        self.__doc__ = func.__doc__
        self._setter = None

    def __set_name__(self, owner: Type, name):
        if not issubclass(owner, Configuration):
            raise ValueError(
                "computed_config_value decorator can only be applied on methods of configuration derivatives")
        self._attr = f"__computed_{name}"
        self._stamp_attr = f"__computed_{name}_stamp"
        self._configuration = owner

    def modifier(self, func: _C1N) -> _C1N:
        self._setter = func
        return func

    def __set__(self, instance, value: _T):
        if instance is None or self._setter is None:
            raise UnsupportedOperationException('immutable field')

        self._setter(instance, value)
        setattr(instance, self._stamp_attr, None)

    def __get__(self, instance, owner) -> _T:
        if instance is None:
            return self

        new_stamp = tuple(id(instance[d]) for d in self._dependency_keys)
        try:
            old_stamp = getattr(instance, self._stamp_attr)
            if old_stamp == new_stamp:
                return getattr(instance, self._attr)
        except AttributeError:
            ...

        new_value = self._func(instance)
        setattr(instance, self._attr, new_value)
        setattr(instance, self._stamp_attr, new_stamp)
        return new_value


def computed_based_on(*based_on_keys: str) -> Callable[[Callable[[Any], _T]], _ComputedConfigValue[_T]]:
    def _computed(func: Callable[[Any], Any]):
        return _ComputedConfigValue(func, based_on_keys)

    return _computed


class FileConfiguration(MutableConfiguration, ABC):
    def __init__(self, *, path: Optional[Path], parent: Optional["Configuration"] = None,
                 data: Optional[Dict[str, Any]] = None):
        super().__init__(parent=parent, data=data)
        self._path = path

    def exists_on_disk(self) -> bool:
        return self._path.exists()

    @abstractmethod
    def generate_content(self) -> str:
        ...

    def save_to(self, path: Optional[Path] = None):
        path = path or self._path
        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(self.generate_content())

    def exists(self) -> bool:
        return self._path.exists()

    @property
    def path(self) -> Path:
        return self._path


class TomlFileConfiguration(FileConfiguration):

    def __init__(self, *, path: Optional[Path], parent: Optional["Configuration"] = None,
                 data: Optional[Dict[str, Any]] = None, dumps: Optional[Callable] = None):
        super().__init__(path=path, parent=parent, data=data)
        self._dumps = dumps

    def generate_content(self) -> str:
        dumps = self._dumps or toml.dumps
        return dumps(self._data)

    @classmethod
    def load(cls, file: Path, parent: Optional[Configuration] = None) -> "TomlFileConfiguration":
        data, dumps = toml.load(file) if file.exists() else ({}, None)
        return cls(path=file, parent=parent, data=data, dumps=dumps)


class InMemConfiguration(MutableConfiguration):

    def save_to(self, path: Optional[Path] = None):
        pass

    @classmethod
    def load(cls, data: Dict[str, Any], parent: Optional[Configuration]) -> "InMemConfiguration":
        return cls(parent=parent, data=data)


class _CaseSensitiveConfigParser(configparser.ConfigParser):
    optionxform = staticmethod(str)


class IniFileConfiguration(FileConfiguration):
    def generate_content(self) -> str:
        class StringWriter:
            def __init__(self):
                self.v: List[str] = []

            def write(self, s: str):
                self.v.append(s)

        sw = StringWriter()
        cp = _CaseSensitiveConfigParser()

        for key_or_section, value_or_content in self._data.items():
            if isinstance(value_or_content, Mapping):
                cp.add_section(key_or_section)
                for key, value in value_or_content.items():
                    cp.set(key_or_section, key, str(value))
            else:
                cp.set("", key_or_section, value_or_content)

        cp.write(sw)
        return ''.join(sw.v)

    @classmethod
    def load(cls, file: Path) -> "IniFileConfiguration":
        data = {}
        if file.exists():
            cp = _CaseSensitiveConfigParser()
            cp.read(str(file))
            for section_name, section_values in cp.items():
                if section_name and section_name != cp.default_section:
                    data[section_name] = {**section_values}
                else:
                    data.update(section_values)

        return cls(path=file, data=data)



from __future__ import annotations

import os
import shutil
from contextlib import contextmanager
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Iterator, List, Dict, Any

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.environments.environment_builder import EnvironmentBuilder
from pkm.api.pkm import pkm
from pkm.api.repositories.repository import Repository
from pkm.config.configuration import TomlFileConfiguration, computed_based_on
from pkm.repositories.shared_pacakges_repo import SharedPackagesRepository
from pkm.utils.commons import NoSuchElementException
from pkm.utils.files import is_relative_to
from pkm.utils.properties import cached_property


class EnvironmentsZoo:

    def __init__(self, cfg: EnvironmentZooConfiguration):
        self.config = cfg
        self.path = cfg.path.parent

    @property
    def _bin_path(self) -> Path:
        return self.path / ".zoo/bin"

    @cached_property
    def attached_repository(self) -> Repository:
        return pkm.repository_loader.load_for_env_zoo(self)

    def create_environment(self, name: str, python: Dependency) -> Environment:

        if (env_path := self.path / name).exists():
            raise FileExistsError(f"environment with name: {name} already exists in the zoo")

        return EnvironmentBuilder.create_matching(env_path, python)

    def delete_environment(self, name: str):
        # shared repository gc
        shutil.rmtree(self.path / name)

    def clean_unused_shared(self):
        repo = self.attached_repository
        if isinstance(repo, SharedPackagesRepository):
            repo.remove_unused_packages(self.list())

    @contextmanager
    def activate(self, env: Dict[str, str] = os.environ):
        prev_path = env.get("PATH")
        new_path_component = str(self._bin_path)
        new_path = f"{new_path_component}{os.pathsep}{prev_path}" if prev_path else new_path_component

        env["PATH"] = new_path
        try:
            yield
        finally:
            if prev_path:
                env["PATH"] = prev_path
            else:
                del env["PATH"]

    def list(self) -> Iterator[Environment]:
        """
        :return: iterator iterating over the requested environments that exists in this zoo
        """
        for e_path in self.path.iterdir():
            if Environment.is_valid(e_path):
                yield Environment(e_path, zoo=self)

    def load_environment(self, name: str) -> Environment:
        if not (e_path := self.path / name).exists():
            raise NoSuchElementException(f"environment named {name} could not be found in this zoo")

        if not Environment.is_valid(e_path):
            raise ValueError(f"directory {e_path} does not contains a valid environment")

        return Environment(e_path, zoo=self)

    def export_script(self, env_name: str, script_name: str):
        env = self.load_environment(env_name)
        env_scripts = Path(env.installation_target.scripts)
        if not (script_file := env_scripts / script_name).exists() \
                and not (script_file := env_scripts / f"{script_name}.exe").exists():
            raise NoSuchElementException(f"script {script_name} could not be find in environment {env_name}")

        shutil.copy(script_file, self._bin_path / script_file.name)

    def export_all_scripts(self, env_name: str, package: str):
        env = self.load_environment(env_name)
        env_scripts = env.installation_target.scripts
        bin_path = self._bin_path

        for file in env.site_packages.installed_package(package).dist_info.installed_files():
            if is_relative_to(file, Path(env_scripts)):
                shutil.copy(file, bin_path / file.name)

    @classmethod
    def load(cls, path: Path) -> EnvironmentsZoo:
        if (cfg_path := path / "env-zoo.toml").exists():
            return EnvironmentsZoo(EnvironmentZooConfiguration.load(cfg_path))

        cfg = EnvironmentZooConfiguration.load(cfg_path)
        cfg.save()

        (path / ".zoo/bin").mkdir(exist_ok=True, parents=True)
        (path / ".zoo/shared").mkdir(exist_ok=True, parents=True)

        return EnvironmentsZoo(cfg)

    @classmethod
    def is_valid(cls, path: Path) -> bool:
        return (path / "env-zoo.toml").exists()


@dataclass(eq=True, frozen=True)
class PackageSharingConfig:
    enabled: bool = True
    exclude: Optional[List[str]] = None

    @classmethod
    def from_config(cls, cfg: Optional[Dict[str, Any]]) -> PackageSharingConfig:
        if not cfg:
            return PackageSharingConfig()

        return PackageSharingConfig(**cfg)

    def to_config(self) -> Dict[str, Any]:
        return copy(self.__dict__)


class EnvironmentZooConfiguration(TomlFileConfiguration):
    package_sharing: PackageSharingConfig

    @computed_based_on("package-sharing")
    def package_sharing(self) -> PackageSharingConfig:
        return PackageSharingConfig(self["package-sharing"])



import os
from pathlib import Path
from typing import List, Optional

from pkm.utils.files import ensure_exists
from pkm.utils.strings import startswith_any


class PthLink:
    """
    taken mostly from the site module documentation:

    A path configuration file is a file whose name has the form name.pth and exists in site packages directories;
    its contents are additional items (one per line) to be added to sys.path.
    Non-existing items are never added to sys.path,
    and no check is made that the item refers to a directory rather than a file.
    No item is added to sys.path more than once.
    Blank lines and lines beginning with # are skipped.
    Lines starting with import (followed by space or tab) are executed.
    """

    def __init__(self, path: Path, links: List[Path], imports: Optional[List[str]] = None):
        """
        :param path: the path to the link itself
        :param links: collection of paths that this link should be responsible for appending to the `sys.path`
        :param imports: collection of imports that this link should be responsible for
                        loading when python sites is resolved
        """
        self.path = path
        self.imports = imports
        self.links = links

    def save(self):
        with self.path.open('w+') as out:
            if self.imports:
                for imp in self.imports:
                    out.write('import ')
                    out.write(imp)
                    out.write(os.linesep)

            for link in self.links:
                out.write(str(link.absolute()))
                out.write('\n')

    @classmethod
    def load(cls, path: Path) -> "PthLink":
        ensure_exists(path)

        imports: List[str] = []
        links: List[Path] = []

        with path.open('r') as pth_fd:
            while line := pth_fd.readline():
                if line := line.strip():
                    if line.startswith("#"):
                        continue
                    elif startswith_any(line, ('import ', 'import\t')):
                        imports.append(line[len("import"):].strip())
                    else:
                        links.append(Path(line))

        return PthLink(path, links, imports)



from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass, replace
from typing import List, Optional

from pkm.api.versions.version import Version, StandardVersion, UrlVersion
from pkm.utils.iterators import distinct
from pkm.utils.sequences import subiter


class VersionSpecifier(ABC):
    """
    represents a version specifier (e.g., >=1.0.1)
    supported operators are: 
    comparison: >,>=,<,<=
    equality: ==, ===, !==, !===, ~=
    url matching: @ 
    """

    def specific_url(self) -> Optional[UrlVersion]:
        """
        :return: if the version specifier is a url matching specifier, returns its version, otherwise returns None  
        """
        return None

    def allows_all(self, other: "VersionSpecifier") -> bool:
        """
        :param other: 
        :return: True if this specifier allows all the versions that are allowed by `other`
        """
        return self.intersect(other) == other

    def allows_any(self, other: "VersionSpecifier") -> bool:
        """
        :param other: 
        :return: True if this specifier allows at least one of the versions that are allowed by `other` 
        """
        return not self.intersect(other).is_none()

    def allows_version(self, version: "Version"):
        """
        :param version: the version to check 
        :return: True if the given `version` is allowed by this version specifier, False otherwise
        """
        return any(segment.allows_version(version) for segment in self._segments())

    def allows_pre_or_dev_releases(self) -> bool:
        """
        :return: True if this specifier accepts version that are pre or dev releases, False otherwise 
        """
        return any(segment.allows_pre_or_dev_releases() for segment in self._segments())

    def intersect(self, other: "VersionSpecifier") -> "VersionSpecifier":
        """
        :param other: 
        :return: a new version specifier that accepts only versions 
                 that are accepted by both self and `other` 
        """
        return _intersect(self, other)

    def union(self, other: "VersionSpecifier") -> "VersionSpecifier":
        """
        :param other: 
        :return: a new version specifier that accepts any versions that are accepted
                 by either self or `other` 
        """
        result = _unite([*self._segments(), *other._segments()])
        return result

    def is_none(self):
        """
        :return: True if this version specifier does not accept any version, False otherwise 
        """
        return self.min == self.max and self.includes_min == self.includes_max == False  # noqa

    def is_any(self):
        """
        :return: True if this version specifier accepts all versions, False otherwise
        """
        return (self.min is self.max is None) and self.includes_min == self.includes_max == True  # noqa

    def _segments(self) -> List["VersionSpecifier"]:
        return [self]

    def difference(self, other: "VersionSpecifier") -> "VersionSpecifier":
        """
        :param other: 
        :return: a new version specifier that accepts any versions that are accepted
                 by self but not by `other` 
        """
        return self.intersect(other.inverse())

    def __lt__(self, other: "VersionSpecifier") -> bool:
        smin, omin = self.min, other.min  # noqa

        if smin == omin:
            if self.includes_min == other.includes_min:
                smax, omax = self.max, other.max

                if smax == omax:
                    if self.includes_max == other.includes_max:
                        return str(self) < str(other)
                    return other.includes_max
                # return Version.less(smax, omax)
                return omax is None or (smax is not None and smax < omax)
            return self.includes_min

        return smin is None or (omin is not None and smin < omin)

    def __le__(self, other: "VersionSpecifier"):
        return other == self or self < other

    @property
    def min(self) -> Optional["Version"]:
        """
        :return: the minimal version that is accepted by this specifier (may-be exclusive)
        """
        return self._segments()[0].min

    @property
    def max(self) -> Optional["Version"]:
        """
        :return: the maximal version that is accepted by this specifier (may-be exclusive)
        """
        return self._segments()[-1].max

    @property
    def includes_min(self) -> bool:
        """
        :return: True if the value returned by `self.min` is inclusive
        """
        return self._segments()[0].includes_min

    @property
    def includes_max(self) -> bool:
        """
        :return: True if the value returned by `self.max` is inclusive
        """
        return self._segments()[1].includes_max

    @abstractmethod
    def inverse(self) -> "VersionSpecifier":
        """
        :return: a new version specifier that accepts any version that was not accepted by this specifier
        """

    @abstractmethod
    def _try_merge(self, other: "VersionSpecifier") -> Optional["VersionSpecifier"]:
        ...

    def __repr__(self):
        return str(self)

    @classmethod
    def parse(cls, txt: str) -> "VersionSpecifier":
        from pkm.api.versions.version_parser import parse_specifier
        return parse_specifier(txt)


@dataclass(frozen=True)
class SpecificVersion(VersionSpecifier):
    version: Version

    def specific_url(self) -> Optional[UrlVersion]:
        if isinstance(self.version, UrlVersion):
            return self.version
        return None

    def allows_version(self, version: "Version"):
        if self.version.is_local():
            return version == self.version
        else:
            return version.without_local() == self.version

    def __str__(self):
        if isinstance(self.version, StandardVersion):
            return f'=={self.version}'
        elif isinstance(self.version, UrlVersion):
            return f'@{self.version}'
        return f"==={self.version}"

    def __repr__(self):
        return str(self)

    def allows_pre_or_dev_releases(self) -> bool:
        return self.version.is_pre_or_dev_release()

    def _try_merge(self, other: "VersionSpecifier") -> Optional["VersionSpecifier"]:
        if other == self:
            return other
        if isinstance(other, SpecificVersion):
            return None
        if isinstance(other, VersionRange):
            if other.min == self.version:
                return replace(other, includes_min=True)
            elif other.max == self.version:
                return replace(other, includes_max=True)
            elif (other.min is None or other.min < self.version) and (other.max is None or self.version < other.max):
                return other
            return None

        assert False, f'merging only support non union versions, attempting to merge: {self} with {other}'
        return None  # noqa

    def inverse(self) -> "VersionSpecifier":
        return VersionUnion([
            VersionRange(max=self.version, includes_max=False),
            VersionRange(min=self.version, includes_min=False)
        ])

    @property
    def min(self) -> Optional["Version"]:
        return self.version

    @property
    def max(self) -> Optional["Version"]:
        return self.version

    @property
    def includes_min(self) -> bool:
        return True

    @property
    def includes_max(self) -> bool:
        return True

    def __lt__(self, other: "VersionSpecifier"):
        if not isinstance(other, SpecificVersion):
            return super().__lt__(other)

        return self.version < other.version


@dataclass(unsafe_hash=True, repr=False)
class VersionRange(VersionSpecifier):
    min: Optional[Version] = None
    max: Optional[Version] = None
    includes_min: bool = None
    includes_max: bool = None

    def __post_init__(self):
        assert self.min is None or self.max is None or self.min <= self.max, f'min > max :: {self.min} > {self.max}'

        self.includes_max = self.includes_max if self.includes_max is not None else self.max is None
        self.includes_min = self.includes_min if self.includes_min is not None else self.min is None

    def allows_pre_or_dev_releases(self) -> bool:
        max_ = self.max
        return max_ is not None and max_.is_pre_or_dev_release()

    # note that pep440 pre-release filtering rules should be implemented in the repository and not here
    def allows_version(self, version: "Version"):

        if self.is_any():
            return not version.is_post_release()
        if self.is_none():
            return False

        min_, max_ = self.min, self.max
        if (min_ is not None and self.includes_min and min_ == version) \
                or (max_ is not None and self.includes_max and max_ == version):
            return True

        if version.is_post_release() and (min_ is None or not min_.is_post_release()):
            return False

        return (min_ is None or min_ < version) and (max_ is None or version < max_)

    def inverse(self) -> "VersionSpecifier":
        new_segments: List["VersionSpecifier"] = []

        if self.is_any():
            return NoVersion
        if self.is_none():
            return AnyVersion

        if self.min is not None:
            new_segments.append(VersionRange(max=self.min, includes_max=not self.includes_min))
        if self.max is not None:
            new_segments.append(VersionRange(min=self.max, includes_min=not self.includes_max))

        return _unite(new_segments)

    def __str__(self):

        if self.is_any():
            return ''
        elif self.is_none():
            return '<none>'

        res = ''
        if self.min is not None:
            res += '>'
            if self.includes_min:
                res += '='
            res += str(self.min)

        if self.max is not None:
            if res:
                res += ', '
            res += '<'
            if self.includes_max:
                res += '='
            res += str(self.max)

        return res

    def _try_merge(self, other: VersionSpecifier) -> Optional[VersionSpecifier]:
        if isinstance(other, SpecificVersion):
            return other._try_merge(self)  # noqa

        min1, min2 = self.min, other.min
        if min2 is not None and (min1 is None or min1 < min2):
            sprv, snxt = self, other
        else:
            sprv, snxt = other, self

        if snxt.min is None or sprv.max is None or snxt.min < sprv.max or \
                (snxt.min == sprv.max and (sprv.includes_max or snxt.includes_min)):
            return VersionRange(min=sprv.min, max=snxt.max, includes_min=sprv.includes_min,
                                includes_max=snxt.includes_max)

        return None


class VersionUnion(VersionSpecifier):
    def __init__(self, constraints: List[VersionSpecifier]):
        self.unions: List[VersionSpecifier] = constraints  # assumed to be ordered..

    def inverse(self) -> "VersionSpecifier":
        result = self.unions[0].inverse()
        for c in subiter(self.unions, 1):
            if result.is_none():
                return result

            result = result.intersect(c.inverse())
        return result

    def _try_merge(self, other: "VersionSpecifier") -> Optional["VersionSpecifier"]:
        assert False, 'merging only support non union versions'
        return None  # noqa

    def __str__(self):
        if len(self.unions) == 2:
            a, b = self.unions
            if a.min is None and b.max is None and a.max == b.min and (a.includes_max is b.includes_min is False):
                return f"!={a.max}"

        return '; '.join(str(it) for it in self.unions)

    def _segments(self) -> List["VersionSpecifier"]:
        return self.unions


NoVersion = VersionRange(includes_min=False, includes_max=False)
AnyVersion = VersionRange(includes_min=True, includes_max=True)


# UTILS

def _unite(segments: List["VersionSpecifier"]) -> "VersionSpecifier":
    if not segments:
        return NoVersion

    segments = sorted(segments)
    new_segments: List[VersionSpecifier] = []

    last: VersionSpecifier = segments[0]
    for segment in subiter(segments, 1):
        joined = last._try_merge(segment)  # noqa
        if joined:
            last = joined
        else:
            new_segments.append(last)
            last = segment
    new_segments.append(last)

    new_segments = list(distinct(new_segments))

    if not new_segments:
        return NoVersion
    elif len(new_segments) == 1:
        return new_segments[0]

    return VersionUnion(new_segments)


def _intersect(a: VersionSpecifier, b: VersionSpecifier) -> "VersionSpecifier":
    swap = False

    if isinstance(a, VersionUnion) or (swap := isinstance(b, VersionUnion)):
        if swap: a, b = b, a  # noqa

        b_segments = [b]
        if isinstance(b, VersionUnion):
            b_segments = b.unions

        return _unite([
            intersection
            for it in a.unions
            for segment in b_segments
            if not (intersection := _intersect(it, segment)).is_none()])

    if isinstance(a, SpecificVersion) or (swap := isinstance(b, SpecificVersion)):
        if swap: a, b = b, a  # noqa

        if isinstance(b, SpecificVersion):
            if a.version == b.version:
                return a
            elif (a.version.is_local() or (swap := b.version.is_local())) \
                    and a.version.without_local() == b.version.without_local():
                return b if swap else a

            return NoVersion

        if isinstance(b, VersionRange):
            if b.is_any() or b.allows_version(a.version.without_local()):
                return a
            return NoVersion

        raise ValueError(f"unexpected version specifier type: {b}")

    # both a and b are version ranges
    if a.is_none() or (swap := b.is_none()):
        return b if swap else a
    if a.is_any() or (swap := b.is_any()):
        return a if swap else b

    min1, min2 = a.min, b.min
    selected_min = min1
    if min1 == min2:
        includes_min = a.includes_min and b.includes_min
    else:
        includes_min = a.includes_min
        if min1 is None or (min2 is not None and min2 > min1):
            selected_min = min2
            includes_min = b.includes_min

    max1, max2 = a.max, b.max
    selected_max = max1
    if max1 == max2:
        includes_max = a.includes_max and b.includes_max
    else:
        includes_max = a.includes_max
        if max1 is None or (max2 is not None and max2 < max1):
            selected_max = max2
            includes_max = b.includes_max

    if includes_max and includes_min and selected_max == selected_min:
        return SpecificVersion(selected_min)
    elif selected_min is None or selected_max is None or selected_min < selected_max:
        return VersionRange(
            min=selected_min, max=selected_max, includes_min=includes_min, includes_max=includes_max)

    return NoVersion



from __future__ import annotations

import re
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, TYPE_CHECKING

from pkm.api.versions.version import Version
from pkm.api.versions.version_specifiers import SpecificVersion
from pkm.utils.commons import UnsupportedOperationException

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.dependencies.dependency import Dependency
    from pkm.api.packages.package_metadata import PackageMetadata
    from pkm.api.packages.package_installation import PackageInstallationTarget


@dataclass(frozen=True)
class PackageDescriptor:
    name: str
    version: Version

    @property
    def expected_src_package_name(self) -> str:
        """
        The expected name of the source package is the same name of the package
         when hyphen is replaced with underscore, there is no guarantee that the
         package author used this name, but this is considered the expected behavior.

        :return: the expected name of the source package that is stored in this package,
        """
        return PackageDescriptor.normalize_src_package_name(self.name)

    def __post_init__(self):
        super().__setattr__('name', PackageDescriptor.normalize_name(self.name).replace('_', '-'))

    def to_dependency(self) -> "Dependency":
        from pkm.api.dependencies.dependency import Dependency
        return Dependency(self.name, SpecificVersion(self.version))

    def write(self) -> Dict[str, Any]:
        return {
            'name': self.name,
            'version': str(self.version),
        }

    @classmethod
    def read(cls, data: Dict[str, Any]) -> "PackageDescriptor":
        return cls(data['name'], Version.parse(data['version']))

    @staticmethod
    def normalize_src_package_name(package_name: str) -> str:
        return PackageDescriptor.normalize_name(package_name).translate({ord('-'): '_', ord('.'): '_'})

    @staticmethod
    def normalize_name(package_name: str) -> str:
        """
        normalize package names (see: https://packaging.python.org/en/latest/specifications/core-metadata/)

        A valid name consists only of ASCII letters and numbers, period, underscore and hyphen.
        It must start and end with a letter or number.
        Distribution names are limited to those which match the following regex (run with re.IGNORECASE):
        `^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$`

        this function replace any non valid chars in the name with '-'
        and then consecutive runs of chars in '-_.' are replaced with a single dash
        finally '-' chars are removed from the start and end of the name

        :param package_name: the package name to normalize
        :return: the normalized name
        """
        if not (result := re.sub("[^A-Z0-9._]+", '-', package_name, flags=re.IGNORECASE).strip('-')):
            raise ValueError(f"empty name after normalization (un-normalized name: '{package_name}')")
        return result

    @classmethod
    def from_dist_name(cls, name: str) -> PackageDescriptor:
        """
        create package descriptor from a distribution file name, assuming standard distribution file name conventions
        :param name: the file name
        :return: the resulted package descriptor
        """
        parts = name.split("-")
        version = Version.parse(parts[1])
        return PackageDescriptor(parts[0], version)


class Package(ABC):

    @property
    @abstractmethod
    def descriptor(self) -> PackageDescriptor:
        """
        descriptor, describing this package
        :return:
        """

    @property
    def published_metadata(self) -> Optional["PackageMetadata"]:
        """
        :return: the package provided metadata,
                 note that this is not the "computed" metadata but instead only the information available to the
                 providing repository about this package
        """
        return None

    @property
    def name(self) -> str:
        return self.descriptor.name

    @property
    def version(self) -> Version:
        return self.descriptor.version

    @abstractmethod
    def _all_dependencies(self, environment: "Environment") -> List["Dependency"]:
        """
        :param environment: the environment that the dependencies should be calculated against
        :return: a list of all the package dependencies (for any environment and extras)
        """

    def dependencies(
            self, environment: "Environment", extras: Optional[List[str]] = None) -> List["Dependency"]:
        """
        :param environment: the environment that the dependencies should be calculated against
        :param extras: the extras to include in the dependencies calculation
        :return: the list of dependencies this package has in order to be installed into the given
        [environment] with the given [extras]
        """

        return [d for d in self._all_dependencies(environment)
                if d.is_applicable_for(environment, extras)]

    @abstractmethod
    def is_compatible_with(self, env: "Environment") -> bool:
        """
        :param env: the environment to check
        :return: true if this package can be installed given its dependencies into the given environment
        """

    @abstractmethod
    def install_to(
            self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
            editable: bool = False):
        """
        installs this package into the given `env`
        :param target: the information about the target to install this package into
        :param user_request: if this package was requested by the user,
               supplying this field will mark the installation as user request
        :param editable: if True and able, the package will be installed in editable mode
        """

    def uninstall(self) -> bool:
        """
        uninstall this package from its package installation target, returns true if the package was removed from the
        site. If a package is a dependency of another 'required' package, then this operation will not remove the
        package but instead will remove the "user-requested" mark from it (and then return False)
        :note: this is an optional operation and is only applicable for "installed" packages
        :return: True if the package was fully removed from the installation target, False otherwise
        """
        raise UnsupportedOperationException("uninstalling is not supported for non installed packages")

    def update_at(self, target: "PackageInstallationTarget", editable: bool = False):
        """
        attempt to update the package from a version installed at the given target to this version
        the update may attempt a full re-installation or a smarted "fast" delta-update like installation
        :param target: the target that contains the package to update
        :param editable: if True, the package will be installed in editable mode
        """
        user_request = False
        if preinstalled := target.site_packages.installed_package(self.name):
            user_request = bool(preinstalled.user_request)
            preinstalled.uninstall()

        self.install_to(target, self.descriptor.to_dependency() if user_request else None, editable=editable)

    def __str__(self):
        return f"{self.name} {self.version}"

    def __repr__(self):
        return f"Package({str(self)})"



import compileall
import re
import warnings
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, TYPE_CHECKING
from zipfile import ZipFile

from pkm.api.distributions.distinfo import DistInfo, RequestedPackageInfo
from pkm.api.distributions.distribution import Distribution
from pkm.api.packages.package import PackageDescriptor
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.versions.version import StandardVersion
from pkm.distributions.executables import Executables
from pkm.utils.archives import extract_archive
from pkm.utils.files import path_to, CopyTransaction, temp_dir

_METADATA_FILE_RX = re.compile("[^/]*\\.dist-info/METADATA")

if TYPE_CHECKING:
    from pkm.api.packages.package import PackageInstallationTarget
    from pkm.api.projects.project import Project
    from pkm.api.environments.environment import Environment


class InstallationException(IOError):
    ...


class WheelDistribution(Distribution):

    def __init__(self, package: PackageDescriptor, wheel: Path):
        self._wheel = wheel
        assert wheel, "no path for wheel provided"
        self._package = package

    def extract_metadata(self, env: Optional["Environment"] = None) -> PackageMetadata:
        with ZipFile(self._wheel) as zipf:
            for name in zipf.namelist():
                if _METADATA_FILE_RX.fullmatch(name):
                    with TemporaryDirectory() as tdir:
                        zipf.extract(name, tdir)
                        return PackageMetadata.load(Path(tdir) / name)
        raise FileNotFoundError("could not find metadata in wheel")

    @property
    def owner_package(self) -> PackageDescriptor:
        return self._package

    def compute_compatibility_tags(self) -> str:
        """
        return the string that represents the compatibility tags in this wheel file name
        :return: the compatibility tag
        """
        return WheelDistribution.compute_compatibility_tags_of(self._wheel)

    @staticmethod
    def compute_compatibility_tags_of(wheel: Path) -> str:
        """
        return the string that represents the compatibility tags in the wheel file name
        :param wheel: the wheel file name
        :return: the compatibility tag
        """
        return '-'.join(wheel.stem.split('-')[-3:])

    @staticmethod
    def expected_wheel_file_name(project: "Project") -> str:
        from pkm.api.environments.environment import Environment
        project_config = project.config.project
        req = project_config.requires_python

        min_interpreter: StandardVersion = req.min \
            if req and not req.is_any() else StandardVersion((Environment.current().interpreter_version.release[0],))

        req_interpreter = 'py' + ''.join(str(it) for it in min_interpreter.release[:2])
        return f"{project.descriptor.expected_src_package_name}-{project.version}-{req_interpreter}-none-any.whl"

    def install_to(self, target: "PackageInstallationTarget", user_request: Optional[RequestedPackageInfo] = None):
        """
        Implementation of wheel installer based on PEP427
        as described in: https://packaging.python.org/en/latest/specifications/binary-distribution-format/
        """
        with temp_dir() as tmp_path:
            extract_archive(self._wheel, tmp_path)
            dist_info = _find_dist_info(tmp_path, self._package)

            wheel_file = dist_info.load_wheel_cfg()
            wheel_file.validate_supported_version()

            entrypoints = dist_info.load_entrypoints_cfg().entrypoints

            site_packages = Path(target.purelib if wheel_file['Root-Is-Purelib'] == 'true' else target.platlib)

            records_file = dist_info.load_record_cfg()
            if not records_file.exists():
                raise InstallationException(
                    f"Unsigned wheel for package {self._package} (no RECORD file found in dist-info)")

            # check that the records hash match
            record_by_path = {r.file: r for r in records_file.records}

            for file in tmp_path.rglob("*"):
                if file.is_dir():
                    continue

                path = str(path_to(tmp_path, file))
                if record := record_by_path.get(path):
                    if not record.hash_signature.validate_against(file):
                        if any(it.name.endswith('.dist-info') for it in file.parents):
                            warnings.warn(f"mismatch hash signature for {file}")
                        else:
                            raise InstallationException(f"File signature not matched for: {record.file}")

                elif file != dist_info.path / "RECORD":
                    raise InstallationException(
                        f"Wheel contains files with no signature in RECORD, "
                        f"e.g., {path}")

            with CopyTransaction() as ct:
                for d in tmp_path.iterdir():
                    if d.is_dir():
                        if d.suffix == '.data':
                            for k in d.iterdir():
                                if not (target_path := getattr(target, k.name, None)):
                                    raise InstallationException(
                                        f'wheel contains data entry with unsupported key: {k.name}')

                                if k.name == 'scripts':
                                    ct.copy_tree(
                                        k, Path(target_path),
                                        file_copy=lambda s, t: Executables.patch_shabang_for_interpreter(
                                            s, t, target.env.interpreter_path))
                                else:
                                    ct.copy_tree(k, Path(target_path))
                        else:
                            ct.copy_tree(d, site_packages / d.name, accept=lambda it: it != records_file.path)
                    else:
                        ct.copy(d, site_packages / d.name)

                # build entry points
                scripts_path = Path(target.scripts)
                for entrypoint in entrypoints:
                    if entrypoint.is_script():
                        ct.touch(Executables.generate_for_entrypoint(target.env, entrypoint, scripts_path))

                # build the new records file
                new_dist_info = DistInfo.load(site_packages / dist_info.path.name)
                new_record_file = new_dist_info.load_record_cfg()

                new_record_file.sign_files(ct.copied_files, site_packages, {
                    r.file: r.hash_signature for r in records_file.records
                })

                new_record_file.save()

                # mark the installer and the requested flag
                (new_dist_info.path / "INSTALLER").write_text(f"pkm\n{self._wheel.name}")
                if user_request:
                    new_dist_info.mark_as_user_requested(user_request)

                # and finally, compile py to pyc
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore')
                    for cc in ct.copied_files:
                        if cc.suffix == '.py':
                            compileall.compile_file(cc, force=True, quiet=2)


def _find_dist_info(unpacked_wheel: Path, package: PackageDescriptor) -> DistInfo:
    dist_info = list(unpacked_wheel.glob("*.dist-info"))
    if not dist_info:
        raise InstallationException(f"wheel for {package} does not contain dist-info")
    if len(dist_info) != 1:
        raise InstallationException(f"wheel for {package} contains more than one possible dist-info")

    return DistInfo.load(dist_info[0])



from typing import Callable, Optional, List, Sequence, TYPE_CHECKING

from pkm.api.versions.version import Version
from pkm.api.versions.version_specifiers import VersionSpecifier
from pkm.utils.parsers import SimpleParser

_EXTRAS_COLLECTION_T = List
_EXTRAS_T = _EXTRAS_COLLECTION_T[str]
_MARKER_PRED_T = Callable[["Environment", _EXTRAS_T], bool]
_MARKER_GET_T = Callable[["Environment", _EXTRAS_T], Sequence[str]]
_MARKER_OP_T = Callable[[Sequence[str], Sequence[str]], bool]

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment


class EnvironmentMarker:

    # noinspection PyShadowingBuiltins
    def __init__(self, expr: str, exec: _MARKER_PRED_T):
        self._expr = expr
        self._exec = exec

    def evaluate_on(self, env: "Environment", extras: _EXTRAS_T) -> bool:
        return self._exec(env, extras)

    def __str__(self) -> str:
        return self._expr

    @classmethod
    def parse_pep508(cls, text: str) -> "EnvironmentMarker":
        return PEP508EnvMarkerParser(text).read_marker()


_MARKER_STR_CHARS = set(
    ' \t0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIZKLMNOPQRSTUVWXYZ().{}-_*#:;,/?[]!`~@$%^&=+|<>')


class PEP508EnvMarkerParser(SimpleParser):

    def _read_identifier(self) -> str:
        if self.peek().isalpha():
            return self.until(lambda i, t: not t[i].isalnum() and t[i] not in '_-.')
        self.raise_err('expecting identifier')

    def _read_str(self) -> str:
        quote = self.match_any("'", "\"")
        oquote = "'" if quote == '"' else '"'

        content = self.until(lambda i, t: t[i] != oquote and t[i] not in _MARKER_STR_CHARS)
        self.match_or_err(quote, f'expecting closing quote: ({quote})')
        return content

    def _read_marker_var(self) -> _MARKER_GET_T:
        if self.peek() in '"\'':
            v_str = self._read_str()
            return lambda e, x: v_str
        else:
            v_ident = self._read_identifier()

            if v_ident == 'extra':
                return lambda e, x: x

            return lambda e, x: e.markers[v_ident]

    def _read_marker_op(self) -> _MARKER_OP_T:
        version_or_string_op = self.match_any("<=", "<", ">=", ">", "!=", "==", "===", "~=")
        if version_or_string_op:
            def _op(a: Sequence[str], b: Sequence[str]) -> bool:

                if version_or_string_op in ('==', '==='):
                    if isinstance(a, _EXTRAS_COLLECTION_T):
                        return b in a
                    elif isinstance(b, _EXTRAS_COLLECTION_T):
                        return a in b

                try:
                    a_ver = Version.parse(str(a))
                    spec = VersionSpecifier.parse(f"{version_or_string_op} {b}")
                    return spec.allows_version(a_ver)
                except ValueError:
                    if version_or_string_op == '~=':
                        raise
                    return eval(f"{repr(str(a))} {version_or_string_op} {repr(str(b))}")

            return _op

        string_op: str = self.match("in") or (self.match("not", "in") and "not in")
        if not string_op:
            self.raise_err("could not parse operator")

        def _op(a: Sequence[str], b: Sequence[str]) -> bool:
            return eval(f"{repr(str(a))} {string_op} {repr(str(b))}")

        return _op

    def _read_marker_expr(self) -> _MARKER_PRED_T:
        self.match_ws()

        if self.match("("):
            result = self._read_marker_compound_expr()
            self.match_or_err(")", "expecting closing paren")
            return result

        left = self._read_marker_var()
        self.match_ws()
        op = self._read_marker_op()
        self.match_ws()
        right = self._read_marker_var()

        return lambda e, x: op(left(e, x), right(e, x))

    def _read_marker_compound_expr(self) -> _MARKER_PRED_T:
        self.match_ws()

        expr: Optional[_MARKER_PRED_T] = None
        self.match_ws()

        ors: List[_MARKER_PRED_T] = []
        while self.is_not_empty() and self.peek() not in ';)':

            if expr:
                lop = self.match_any('and', 'or')
                if not lop:
                    self.raise_err('expecting logical operator: and/or')

                next_expr = self._read_marker_expr()

                if lop == 'and':
                    def expr(e, x, ex1=expr, ex2=next_expr):
                        return ex1(e, x) and ex2(e, x)

                else:
                    ors.append(expr)
                    expr = next_expr

            else:
                expr = self._read_marker_expr()

            self.match_ws()

        if ors:
            ors.append(expr)

            def expr(e, x):
                return any(expr_(e, x) for expr_ in ors)

        if expr is None:
            self.raise_err('expecting environment marker expression')

        return expr

    def read_marker(self) -> EnvironmentMarker:
        p = self.position
        exe = self._read_marker_compound_expr()
        return EnvironmentMarker(self.text[p:self.position], exe)
